import subprocess
import warnings
from ast import literal_eval

from numpy import mean, std
from pandas import read_csv, unique, DataFrame

import settings

from plotly import graph_objects as go

from authentication.results_generator import ResultsGenerator
from helpers.helper import Helper


class ConfidenceLevel(Helper):
    def __init__(self, datasets=None, results=None):
        super().__init__()
        self.datasets = datasets
        self.results = results

        self.name = "ca_confidence"
        self.gnuplot_script = "ca_confidence.plt"

        warnings.filterwarnings("ignore")

    def generate(self, cache=True):
        """
        Generate the confidence levels for 10 seconds lookback window.
        """
        if cache:
            self.results = self._check_for_results()
            if self.results:
                print("Using cached results.")
                return
        generator = ResultsGenerator(self.datasets, [10], [1, 2, 3], self.name)
        self.results = generator.generate()

    def visualize(self, show=True):
        """
        Visualize the resulting csv file and generate a gnuplot input csv file.
        """
        df = read_csv(self.results)

        TASKS = list(unique(df["task"]))
        window = 10

        results = {
            "user": [],
            "task": [],
            "true": [],
            "true_sd": [],
            "false": [],
            "false_sd": [],
        }
        for task in TASKS:
            sub = df[(df["task"] == task) & (df["window"] == window)]
            users = [x for x in literal_eval(list(sub["user_confidences"])[0])]

            accuracies = literal_eval(list(sub["user_accuracies"])[0])
            confidences = literal_eval(list(sub["user_confidences"])[0])
            for user in users:
                accu = accuracies[user]
                conf = confidences[user]
                true = []
                false = []
                for i, acc in enumerate(accu):
                    if acc:
                        true.append(conf[i])
                    else:
                        false.append(conf[i])
                results["user"].append(user)
                results["task"].append(task)
                results["true"].append(mean(true))
                results["true_sd"].append(std(true))
                results["false"].append(mean(false))
                results["false_sd"].append(std(false))

        results = DataFrame(results)
        users = list(unique(results["user"]))

        final_results = {
            "user": [],
            "true_avg": [],
            "true_std": [],
            "false_avg": [],
            "false_std": [],
        }
        for user in users:
            sub = results[results["user"] == user]
            final_results["user"].append(user)
            final_results["true_avg"].append(mean(sub["true"]))
            final_results["true_std"].append(mean(sub["true_sd"]))
            final_results["false_avg"].append(mean(sub["false"]))
            final_results["false_std"].append(mean(sub["false_sd"]))

        graphs = [
            go.Bar(
                x=final_results["user"],
                y=final_results["true_avg"],
                name="True",
                error_y=(
                    {
                        "type": "data",
                        "array": final_results["true_std"],
                        "visible": True,
                    }
                ),
            ),
            go.Bar(
                x=final_results["user"],
                y=final_results["false_avg"],
                name="False",
                error_y=(
                    {
                        "type": "data",
                        "array": final_results["false_std"],
                        "visible": True,
                    }
                ),
            ),
        ]
        fig = go.Figure(data=graphs)
        fig.update_layout(
            #         title="Per user true and false confidence level.".format(),
            xaxis_title="User",
            yaxis_title="Confidence level",
            xaxis_type="category",
            yaxis_range=[0, 1.1],
        )

        if show:
            fig.show()

        df = DataFrame(final_results)
        with open(settings.VISUALIZATION_DIR + "ca_confidence.csv", "w") as csv:
            csv.write("User\tTrue\tFalse\tTrue\tFalse\n")
            for user in users:
                sub = df[df["user"] == user]
                pos_acc = sub["true_avg"].values[0]
                pos_sd = sub["true_std"].values[0]
                neg_acc = sub["false_avg"].values[0]
                neg_sd = sub["false_std"].values[0]
                csv.write(
                    "{}\t{}\t{}\t{}\t{}\n".format(
                        user, pos_acc, neg_acc, pos_sd, neg_sd
                    )
                )
