import subprocess
from os import listdir, mkdir
from os.path import isdir

from IPython.core.display import clear_output
from numpy import argsort, mean, arange, std
from pandas import unique, concat, DataFrame, read_csv
from plotly import graph_objects as go
from sklearn.ensemble import RandomForestClassifier
from sklearn.feature_selection import SelectKBest, mutual_info_classif

import settings
from helpers.helper import Helper, gini


class ContinuousAuthentication(Helper):
    def __init__(
        self, datasets, intrusions_dir="intrusions/", model=RandomForestClassifier
    ):
        super().__init__()

        self.datasets = datasets
        self.intrusions_dir = intrusions_dir

        self._check_intrusions_dir()

        self.model_name = model.__name__
        if self.model_name == "RandomForestClassifier":
            self.model = model(n_estimators=100)
        elif self.model_name == "GaussianNB":
            self.model = model(var_smoothing=1e-9)
        elif self.model_name == "LogisticRegression":
            self.model = model(multi_class="multinomial", n_jobs=8)
        else:
            raise Exception("Unknown model attribute.")

    def simulate_breach(self):
        """
        Simulate an credentials breach by combining data of two
        different users and use all others as legitimate users.
        """
        for trial in [1, 2]:
            # Get already obtained results, to not calculate them again
            already_processed = []
            for x in [
                x
                for x in listdir(settings.RESULTS_DIR + self.intrusions_dir)
                if x.split("_")[3][3:] == "csv"
                and x.split("_")[1] == str(trial).zfill(2)
                and x.split("_")[0] == self.model_name
            ]:
                legit, attack = x.split("_")[2:4]
                attack = attack[:2]
                already_processed.append((int(legit), int(attack)))

            task = trial
            segment_interval = 1
            data = self.datasets[
                "{}_{}".format(str(task).zfill(2), str(segment_interval).zfill(3))
            ]
            # How many time steps back we take to calculate mean conf
            lookback = 5
            # Skip this many time frames
            time_step = 1
            # Take this many last data rows of every user
            window = 5
            # In which iteration we start
            start = 3

            # Iterate through pairs of legit and attack users
            u = list(unique(data["user"]))
            for legit in u:
                for attack in u:
                    # Skip pairs for which we already have data
                    if legit == attack or any(
                        [
                            True if legit == x[0] and attack == x[1] else False
                            for x in already_processed
                        ]
                    ):
                        continue
                    data = self.datasets[
                        "{}_{}".format(
                            str(task).zfill(2), str(segment_interval).zfill(3)
                        )
                    ]

                    # Users
                    seances = self._get_users_data(data)
                    users = list(unique(data["user"]))
                    users.remove(legit)
                    users.remove(attack)

                    # Construct the attack set, label all attack data with legit user
                    legit_d = data[data["seance"] == seances[legit][0]]
                    attack_d = data[data["seance"] == seances[attack][0]]
                    breaking_point = min(len(legit_d), len(attack_d)) // 2
                    attack_set = concat(
                        [
                            legit_d.iloc[:breaking_point, :],
                            attack_d.iloc[breaking_point:, :],
                        ]
                    )
                    real_attack_user = list(attack_set["user"])
                    attack_set["user"] = [legit for _ in range(len(attack_set))]

                    # Subset users
                    data = data[data.user.isin(users)]

                    # Split the data
                    train_seances = [
                        x[trial - 1] for x in self._get_users_data(data).values()
                    ]
                    training_set = concat(
                        [data[data["seance"].isin(train_seances)], attack_set]
                    )

                    # print(breaking_point)
                    # print(train_seances)
                    # print(list(unique(training_set["user"])))

                    seances = list(set(training_set["seance"]))

                    # results dict
                    results = {
                        "time": [],
                        "confidence": [],
                        "avg_5_conf": [],
                        "avg_10_conf": [],
                        "avg_20_conf": [],
                        "avg_40_conf": [],
                        "real_user": [],
                        "predicted": [],
                        "user_count": [],
                    }

                    # Iterate through time and detect breaches
                    for i in range(start, len(attack_set) - 1, time_step):
                        # Subset data, according to the current time step
                        train = []
                        for seance in seances:
                            train.append(
                                training_set[training_set["seance"] == seance].iloc[
                                    max(0, i - window) : i, :
                                ]
                            )

                        train = concat(train)
                        test = attack_set.iloc[i : i + 1, :]

                        # Feature selection
                        X = train.iloc[:, 2:].values
                        y = train.iloc[:, 0].values.ravel()
                        cols = list(data.columns)
                        f_cols = list(train.columns)
                        sel = SelectKBest(mutual_info_classif, k=2)
                        try:
                            sel.fit(X, y)
                        except ValueError:
                            print("OUT OF DATA")
                            break
                        indices = argsort(sel.scores_)
                        feature_set = cols[0:2] + [f_cols[x] for x in indices[-10:]]

                        # Take the selected features
                        X_train = train[feature_set].iloc[:, 2:].values
                        y_train = train[feature_set].iloc[:, 0].values.ravel()
                        X_test = test[feature_set].iloc[:, 2:].values
                        y_test = test[feature_set].iloc[:, 0].values.ravel()

                        # Fit the model
                        # Break the loop if there is only a single class to predict to left
                        if len(unique(y_train)) <= 1:
                            break
                        self.model.fit(X_train, y_train)

                        # Predict the classes
                        y_pred = self.model.predict(X_test)
                        y_prob = self.model.predict_proba(X_test)

                        # Get the user index for the confidence
                        ui = list(self.model.classes_).index(y_pred[0])

                        # Calculate the mean of last n confidence values
                        curr_5_value = mean(results["confidence"][-lookback:])
                        curr_10_value = mean(results["confidence"][-(lookback * 2) :])
                        curr_20_value = mean(results["confidence"][-(lookback * 4) :])
                        curr_40_value = mean(results["confidence"][-(lookback * 8) :])

                        clear_output()
                        print("time: {}".format(i))
                        print("Legit: {}, attacker: {}".format(legit, attack))

                        # Append to the results
                        results["time"].append(i * segment_interval)
                        results["confidence"].append(y_prob[0][ui])
                        results["avg_5_conf"].append(curr_5_value)
                        results["avg_10_conf"].append(curr_10_value)
                        results["avg_20_conf"].append(curr_20_value)
                        results["avg_40_conf"].append(curr_40_value)
                        results["real_user"].append(real_attack_user[i])
                        results["predicted"].append(y_pred[0])
                        results["user_count"].append(len(list(unique(train["user"]))))

                    # Generate the csv of the final state
                    DataFrame(results).to_csv(
                        settings.RESULTS_DIR
                        + self.intrusions_dir
                        + "{}_{}_{}_{}.csv".format(
                            self.model_name,
                            str(trial).zfill(2),
                            str(legit).zfill(2),
                            str(attack).zfill(2),
                        )
                    )

    def visualize(self, show=True):
        """
        Visualize delay and frr vs threshold for different lookback windows.
        """
        # users = [[7, 8, 10, 11, 13, 14, 16, 17, 18, 20, 21, 22, 23, 24, 25, 26, 27]]
        users = [[7, 8, 10, 13, 23, 25, 26, 27]]
        for user in users:
            for trial in [1]:
                results = {
                    "threshold": [],
                    "lookback": [],
                    "frr": [],
                    "far": [],
                    "frr_sd": [],
                    "frr_gini": [],
                    "delay": [],
                    "delay_sd": [],
                }
                files = sorted(
                    [
                        x
                        for x in listdir(settings.RESULTS_DIR + self.intrusions_dir)
                        if x[-3:] == "csv"
                        and x.split("_")[0] == self.model_name
                        and x.split("_")[1] == str(trial).zfill(2)
                        # For a single user at a time
                        # and x.split("_")[2] == str(user).zfill(2)
                        # Subsamling users
                        and x.split("_")[2] in [str(y).zfill(2) for y in user]
                        # and x.split("_")[3][:2] in [str(y).zfill(2) for y in user]
                    ]
                )

                files = [
                    read_csv(settings.RESULTS_DIR + self.intrusions_dir + file).iloc[
                        :, 1:
                    ]
                    for file in files
                ]

                THRESHOLDS = list(arange(0, 1, 0.00625))
                # THRESHOLDS = list(arange(0.425, 0.75, 0.00625))
                LOOKBACKS = [5, 10, 20, 40]
                for conf in LOOKBACKS:
                    for threshold in THRESHOLDS:
                        false_rejection_rate = []
                        false_acceptance_rate = []
                        delay = []

                        # Calculate some per run statistics
                        for df in files:
                            # print(df)
                            # print("=====================================")
                            legit = df["real_user"].head(1).values[0]
                            attack = df["real_user"].tail(1).values[0]

                            # Calculate FRR
                            legit_part = df[df["real_user"] == legit]

                            legit_len = len(legit_part)
                            false_neg = len(
                                legit_part[
                                    legit_part["avg_{}_conf".format(conf)] <= threshold
                                ]
                            )
                            frr = false_neg / legit_len
                            false_rejection_rate.append(frr)

                            # Calculate FAR
                            attack_part = df[df["real_user"] == attack]

                            attack_len = len(attack_part)
                            false_pos = len(
                                attack_part[
                                    attack_part["avg_{}_conf".format(conf)] > threshold
                                ]
                            )

                            far = false_pos / attack_len
                            false_acceptance_rate.append(far)
                            # print(false_pos)
                            # print(attack_len)
                            # print(far)
                            # print("-------------------------------------")

                            # Calculate attack detection delay
                            attack_part = df[df["real_user"] == attack]
                            attack_start = int(attack_part.iloc[0, :]["time"])
                            try:
                                detection = int(
                                    attack_part[
                                        attack_part["avg_{}_conf".format(conf)]
                                        <= threshold
                                    ].iloc[0, :]["time"]
                                )
                            except IndexError:
                                continue

                            # print(attack_start)
                            # print(detection)
                            # print(detection - attack_start)
                            # print("=====================================")

                            delay.append(detection - attack_start)

                        results["threshold"].append(round(threshold, 3))
                        results["lookback"].append(conf)
                        results["frr"].append(mean(false_rejection_rate))
                        results["far"].append(mean(false_acceptance_rate))
                        results["frr_sd"].append(std(false_rejection_rate))
                        results["frr_gini"].append(gini(sorted(false_rejection_rate)))
                        results["delay"].append(mean(delay))
                        results["delay_sd"].append(std(delay))

                df = DataFrame(results)
                with open(
                    settings.VISUALIZATION_DIR + self.model_name + "_delay.csv", "w"
                ) as csv:
                    for threshold in THRESHOLDS:
                        threshold = round(threshold, 3)
                        csv.write("{}\t".format(threshold))
                        for lookback in LOOKBACKS:
                            csv.write(
                                "{}\t".format(
                                    list(
                                        df[
                                            (df["threshold"] == threshold)
                                            & (df["lookback"] == lookback)
                                        ]["delay"].values
                                    )[0]
                                )
                            )
                        csv.write("\n")

                with open(
                    settings.VISUALIZATION_DIR + self.model_name + "_frr.csv", "w"
                ) as csv:
                    for threshold in THRESHOLDS:
                        threshold = round(threshold, 3)
                        csv.write("{}\t".format(threshold))
                        for lookback in LOOKBACKS:
                            csv.write(
                                "{}\t".format(
                                    list(
                                        df[
                                            (df["threshold"] == threshold)
                                            & (df["lookback"] == lookback)
                                        ]["frr"].values
                                    )[0]
                                )
                            )

                        for lookback in LOOKBACKS:
                            csv.write(
                                "{}\t".format(
                                    list(
                                        df[
                                            (df["threshold"] == threshold)
                                            & (df["lookback"] == lookback)
                                        ]["far"].values
                                    )[0]
                                )
                            )
                        csv.write("\n")

                if show:
                    # FRR
                    graphs = []
                    for lookback in LOOKBACKS:
                        sub = df[df["lookback"] == lookback]
                        graphs.append(
                            go.Scatter(
                                x=sub["threshold"],
                                y=sub["frr"],
                                name="Lookback {}".format(lookback),
                            )
                        )
                    fig = go.Figure(data=graphs)
                    fig.update_layout(
                        title="Frr for different threshold and lookback intervals.",
                        xaxis_title="Threshold",
                        xaxis_type="category",
                        yaxis_title="False rejection rate",
                    )
                    # fig.show()

                    # FAR
                    graphs = []
                    for lookback in LOOKBACKS:
                        sub = df[df["lookback"] == lookback]
                        graphs.append(
                            go.Scatter(
                                x=sub["threshold"],
                                y=sub["far"],
                                name="Lookback {}".format(lookback),
                            )
                        )
                    fig = go.Figure(data=graphs)
                    fig.update_layout(
                        title="Far for different threshold and lookback intervals.",
                        xaxis_title="Threshold",
                        xaxis_type="category",
                        yaxis_title="False acceptance rate",
                    )
                    # fig.show()

                    # FRR and FAR
                    graphs = []
                    for lookback in LOOKBACKS:
                        sub = df[df["lookback"] == lookback]
                        graphs.append(
                            go.Scatter(
                                x=sub["threshold"],
                                y=sub["frr"],
                                name="Lookback {} frr".format(lookback),
                            )
                        )
                        graphs.append(
                            go.Scatter(
                                x=sub["threshold"],
                                y=sub["far"],
                                name="Lookback {} far".format(lookback),
                            )
                        )
                        sub = sub.reset_index()
                        eer_indx = (sub["frr"] - sub["far"]).abs().idxmin()
                        print("Threshold: {}".format(THRESHOLDS[eer_indx]))
                        print(sub[["frr", "far"]].iloc[eer_indx, :].mean(axis=0))
                        print()
                    fig = go.Figure(data=graphs)
                    fig.update_layout(
                        title="Far and frr for different threshold and lookback intervals.",
                        xaxis_title="Threshold",
                        xaxis_type="category",
                        yaxis_title="False acceptance and rejection rates",
                    )
                    fig.show()

                    # Delay
                    graphs = []
                    for lookback in LOOKBACKS:
                        sub = df[df["lookback"] == lookback]
                        graphs.append(
                            go.Scatter(
                                x=sub["threshold"],
                                y=sub["delay"],
                                #             error_y={"array": sub["delay_sd"]},
                                name="Lookback {}".format(lookback),
                            )
                        )
                    fig = go.Figure(data=graphs)
                    fig.update_layout(
                        title="Delay for different threshold and lookback intervals.",
                        xaxis_title="Threshold",
                        xaxis_type="category",
                        yaxis_title="Delay [s]",
                    )
                    fig.show()

                    # Gini indices
                    graphs = []
                    for lookback in LOOKBACKS:
                        sub = df[df["lookback"] == lookback]
                        graphs.append(
                            go.Scatter(
                                x=sub["threshold"],
                                y=sub["frr_gini"],
                                #             error_y={"array": sub["delay_sd"]},
                                name="Lookback {}".format(lookback),
                            )
                        )
                    fig = go.Figure(data=graphs)
                    fig.update_layout(
                        title="Gini index of frr for different threshold and lookback intervals.",
                        xaxis_title="Threshold",
                        xaxis_type="category",
                        yaxis_title="Gini index",
                    )
                    # fig.show()
                    # print(
                    #     "COMMENT: Gini index seems to drop with the ascending number of non-zero frr values."
                    # )
        return df

    def visualize_nb(self, show=True):
        """
        Visualize delay and frr vs threshold for different lookback windows.
        """
        results = {
            "threshold": [],
            "lookback": [],
            "frr": [],
            "far": [],
            "frr_sd": [],
            "frr_gini": [],
            "delay": [],
            "delay_sd": [],
        }
        for trial in [1]:
            files = sorted(
                [
                    x
                    for x in listdir(settings.RESULTS_DIR + self.intrusions_dir)
                    if x[-3:] == "csv"
                    and x.split("_")[0] == self.model_name
                    and x.split("_")[1] == str(trial).zfill(2)
                ]
            )[:3]

            files = [
                read_csv(settings.RESULTS_DIR + self.intrusions_dir + file).iloc[:, 1:]
                for file in files
            ]

            false_rejection_rate = []
            false_acceptance_rate = []
            delay = []

            # Calculate some per run statistics
            for df in files:
                print(df)
                legit = df["real_user"].head(1).values[0]
                attack = df["real_user"].tail(1).values[0]

                # Calculate FRR
                legit_part = df[df["real_user"] == legit]

                legit_len = len(legit_part)
                false_neg = len(legit_part[legit_part["predicted"] != legit])
                frr = false_neg / legit_len
                false_rejection_rate.append(frr)

                # Calculate FAR
                attack_part = df[df["real_user"] == attack]

                attack_len = len(attack_part)
                false_pos = len(attack_part[attack_part["predicted"] == legit])

                far = false_pos / attack_len
                false_acceptance_rate.append(far)

                # Calculate attack detection delay
                attack_part = df[df["real_user"] == attack]
                attack_start = int(attack_part.iloc[0, :]["time"])
                try:
                    detection = int(
                        attack_part[attack_part["predicted"] != legit].iloc[0, :][
                            "time"
                        ]
                    )
                except IndexError:
                    continue

                delay.append(detection - attack_start)
            print(mean(false_rejection_rate))
            print(mean(false_acceptance_rate))
            print(mean(delay))

            # results["threshold"].append(round(threshold, 3))
            # results["lookback"].append(conf)
            # results["frr"].append(mean(false_rejection_rate))
            # results["far"].append(mean(false_acceptance_rate))
            # results["frr_sd"].append(std(false_rejection_rate))
            # results["frr_gini"].append(gini(sorted(false_rejection_rate)))
            # results["delay"].append(mean(delay))
            # results["delay_sd"].append(std(delay))

        df = DataFrame(results)
        with open(
            settings.VISUALIZATION_DIR + self.model_name + "_delay.csv", "w"
        ) as csv:
            for threshold in THRESHOLDS:
                threshold = round(threshold, 3)
                csv.write("{}\t".format(threshold))
                for lookback in LOOKBACKS:
                    csv.write(
                        "{}\t".format(
                            list(
                                df[
                                    (df["threshold"] == threshold)
                                    & (df["lookback"] == lookback)
                                ]["delay"].values
                            )[0]
                        )
                    )
                csv.write("\n")

        with open(
            settings.VISUALIZATION_DIR + self.model_name + "_frr.csv", "w"
        ) as csv:
            for threshold in THRESHOLDS:
                threshold = round(threshold, 3)
                csv.write("{}\t".format(threshold))
                for lookback in LOOKBACKS:
                    csv.write(
                        "{}\t".format(
                            list(
                                df[
                                    (df["threshold"] == threshold)
                                    & (df["lookback"] == lookback)
                                ]["frr"].values
                            )[0]
                        )
                    )
                csv.write("\n")

        if show:
            # FRR
            graphs = []
            for lookback in LOOKBACKS:
                sub = df[df["lookback"] == lookback]
                graphs.append(
                    go.Scatter(
                        x=sub["threshold"],
                        y=sub["frr"],
                        name="Lookback {} frr".format(lookback),
                    )
                )
            fig = go.Figure(data=graphs)
            fig.update_layout(
                title="Frr for different threshold and lookback intervals.",
                xaxis_title="Threshold",
                xaxis_type="category",
                yaxis_title="False rejection rate",
            )
            fig.show()

            # FAR
            graphs = []
            for lookback in LOOKBACKS:
                sub = df[df["lookback"] == lookback]
                graphs.append(
                    go.Scatter(
                        x=sub["threshold"],
                        y=sub["far"],
                        name="Lookback {} far".format(lookback),
                    )
                )
            fig = go.Figure(data=graphs)
            fig.update_layout(
                title="Far for different threshold and lookback intervals.",
                xaxis_title="Threshold",
                xaxis_type="category",
                yaxis_title="False acceptance rate",
            )
            fig.show()

            # FRR
            graphs = []
            for lookback in LOOKBACKS:
                sub = df[df["lookback"] == lookback]
                graphs.append(
                    go.Scatter(
                        x=sub["threshold"],
                        y=sub["frr"],
                        name="Lookback {} frr".format(lookback),
                    )
                )
            fig = go.Figure(data=graphs)
            fig.update_layout(
                title="Frr for different threshold and lookback intervals.",
                xaxis_title="Threshold",
                xaxis_type="category",
                yaxis_title="False rejection rate",
            )
            fig.show()

            # FRR and FAR
            graphs = []
            for lookback in LOOKBACKS:
                sub = df[df["lookback"] == lookback]
                graphs.append(
                    go.Scatter(
                        x=sub["threshold"],
                        y=sub["far"],
                        name="Lookback {} far".format(lookback),
                    )
                )
                graphs.append(
                    go.Scatter(
                        x=sub["threshold"],
                        y=sub["frr"],
                        name="Lookback {} frr".format(lookback),
                    )
                )
            fig = go.Figure(data=graphs)
            fig.update_layout(
                title="Far for different threshold and lookback intervals.",
                xaxis_title="Threshold",
                xaxis_type="category",
                yaxis_title="False acceptance rate",
            )
            fig.show()

            # Delay
            graphs = []
            for lookback in LOOKBACKS:
                sub = df[df["lookback"] == lookback]
                graphs.append(
                    go.Scatter(
                        x=sub["threshold"],
                        y=sub["delay"],
                        #             error_y={"array": sub["delay_sd"]},
                        name="Lookback {}".format(lookback),
                    )
                )
            fig = go.Figure(data=graphs)
            fig.update_layout(
                title="Delay for different threshold and lookback intervals.",
                xaxis_title="Threshold",
                xaxis_type="category",
                yaxis_title="Delay [s]",
            )
            fig.show()

            # Gini indices
            graphs = []
            for lookback in LOOKBACKS:
                sub = df[df["lookback"] == lookback]
                graphs.append(
                    go.Scatter(
                        x=sub["threshold"],
                        y=sub["frr_gini"],
                        #             error_y={"array": sub["delay_sd"]},
                        name="Lookback {}".format(lookback),
                    )
                )
            fig = go.Figure(data=graphs)
            fig.update_layout(
                title="Gini index of frr for different threshold and lookback intervals.",
                xaxis_title="Threshold",
                xaxis_type="category",
                yaxis_title="Gini index",
            )
            fig.show()
            print(
                "COMMENT: Gini index seems to drop with the ascending number of non-zero frr values."
            )
        return df

    def generate_multiple_eps(self):
        """
        Generate the eps graph using gnuplot.
        """
        self.gnuplot_script = "delay.plt"
        self.generate_eps()
        self.gnuplot_script = "frr.plt"
        self.generate_eps()

    @staticmethod
    def _get_users_data(data):
        """
        Get seance id for each user in a form of a dict.
        """
        users = {}
        for user in list(set(data["user"])):
            x = data[data["user"] == user]
            users.update({user: sorted(list(set(x["seance"])))})
        return users

    def _check_intrusions_dir(self):
        """
        Check if intrusions dir exists and create it if it doesn't.
        """
        if not isdir(settings.RESULTS_DIR + self.intrusions_dir):
            mkdir(settings.RESULTS_DIR + self.intrusions_dir)
