import subprocess
import warnings
from ast import literal_eval

from numpy import mean, std
from pandas import read_csv, unique, DataFrame

import settings
from authentication.results_generator import ResultsGenerator
from plotly import graph_objects as go

from helpers.helper import Helper


class LookbackWindow(Helper):
    def __init__(self, datasets=None, results=None):
        super().__init__()
        self.datasets = datasets
        self.results = results

        self.name = "lookback"
        self.gnuplot_script = "ca_accuracy.plt"

        warnings.filterwarnings("ignore")

    def generate(self, cache=True):
        """
        Generate the results with various lookback windows.
        Optionally check for already calculated results.
        """
        if cache:
            self.results = self._check_for_results()
            if self.results:
                print("Using cached results.")
                return
        generator = ResultsGenerator(
            self.datasets, [10, 25, 50, 100], [1, 2, 3], self.name
        )
        self.results = generator.generate()

    def visualize(self, show=True):
        """
        Visualize the resulting csv file and generate a gnuplot input csv file.
        """
        df = read_csv(self.results)

        TASKS = list(unique(df["task"]))
        WINDOWS = list(unique(df["window"]))

        results = {"task": [], "window": [], "avg": [], "std": []}
        for task in TASKS:
            for window in WINDOWS:
                sub = df[(df["task"] == task) & (df["window"] == window)]
                results["avg"].append(mean(literal_eval(list(sub["accuracy"])[0])))
                results["std"].append(std(literal_eval(list(sub["accuracy"])[0])))
                results["task"].append(task)
                results["window"].append(window)

        results = DataFrame(results)
        graphs = []
        for window in WINDOWS:
            sub = results[results["window"] == window]
            graphs.append(
                go.Bar(
                    x=sub["task"],
                    y=sub["avg"],
                    name="Window size {}s".format(window),
                    error_y={"array": sub["std"]},
                )
            )
        fig = go.Figure(data=graphs)
        fig.update_layout(
            title="Mean user accuracy for different window size",
            xaxis_title="Task",
            yaxis_title="Classification accuracy",
        )
        #     fig.update_yaxes(range=[0.8,1.1])
        if show:
            fig.show()

        with open(settings.VISUALIZATION_DIR + "ca_accuracy.csv", "w") as csv:
            csv.write("\t")
            for window in WINDOWS:
                csv.write(str(window) + "s\t")
            for window in WINDOWS:
                csv.write(str(window) + "s\t")
            csv.write("\n")
            for task in TASKS:
                csv.write(str(task) + "\t")
                for window in WINDOWS:
                    acc = str(
                        results[
                            (results["task"] == task) & (results["window"] == window)
                        ]["avg"].values[0]
                    )
                    csv.write(acc + "\t")
                for window in WINDOWS:
                    sd = str(
                        results[
                            (results["task"] == task) & (results["window"] == window)
                        ]["std"].values[0]
                    )
                    csv.write(sd + "\t")
                csv.write("\n")
