import warnings
from datetime import datetime
from os import mkdir
from os.path import isdir

from IPython.core.display import clear_output
from numpy import argsort
from pandas import unique, DataFrame, concat
from sklearn import metrics
from sklearn.ensemble import RandomForestClassifier
from sklearn.feature_selection import SelectKBest, mutual_info_classif

import settings


class ResultsGenerator:
    def __init__(
        self, datasets, windows, tasks, name,
    ):

        self.datasets = datasets
        self.windows = windows
        self.tasks = tasks
        self.name = name

    def generate(self):
        """
        Generate the results, based on the given parameters.
        """
        warnings.filterwarnings("ignore")
        # Generate results
        final_results = {
            "task": [],
            "window": [],
            "time": [],
            "accuracy": [],
            "user_accuracies": [],
            "user_confidences": [],
        }
        for task in self.tasks:
            for window in self.windows:
                # Data generation and cache
                segment_interval = 1
                data = self.datasets[
                    "{}_{}".format(str(task).zfill(2), str(segment_interval).zfill(3))
                ]

                # Users
                users = list(unique(data["user"]))

                # Subset users
                data = data[data.user.isin(users)]

                # Split the data
                seances = self._get_users_data(data)
                train_seances = [x[0] for x in seances.values()]
                training_set = data[data["seance"].isin(train_seances)]
                seances = list(set(training_set["seance"]))

                i = 2
                time_step = 1
                user_accuracies = {}
                user_confidences = {}
                for user in users:
                    user_accuracies.update({user: []})
                    user_confidences.update({user: []})
                results = {
                    "time": [],
                    "accuracy": [],
                    "user_accuracies": user_accuracies,
                    "user_confidences": user_confidences,
                }
                while True:
                    clear_output()
                    print(self.tasks, self.windows)
                    print(task, window)
                    print(i)

                    # Subset data, according to the current time step
                    train = []
                    test = []
                    for seance in seances:
                        train.append(
                            training_set[training_set["seance"] == seance].iloc[
                                max(0, i - window) : i, :
                            ]
                        )
                        test.append(
                            training_set[training_set["seance"] == seance].iloc[
                                i : i + 1, :
                            ]
                        )
                    train = concat(train)
                    print(train.size)
                    test = concat(test)
                    if test.empty:
                        print("RUN OUT OF DATA.")
                        break

                    # Feature selection
                    X = train.iloc[:, 2:].values
                    y = train.iloc[:, 0].values.ravel()
                    cols = list(train.columns)
                    f_cols = list(train.iloc[:, 2:].columns)
                    sel = SelectKBest(mutual_info_classif, k=2)
                    sel.fit(X, y)
                    indices = argsort(sel.scores_)
                    feature_set = cols[0:2] + [f_cols[x] for x in indices[-10:]]

                    # Take the selected features
                    X_train = train[feature_set].iloc[:, 2:].values
                    y_train = train[feature_set].iloc[:, 0].values.ravel()
                    X_test = test[feature_set].iloc[:, 2:].values
                    y_test = test[feature_set].iloc[:, 0].values.ravel()

                    if not len(X_test):
                        print("RUN OUT OF DATA.")
                        break

                    # Fit the model
                    classifier = RandomForestClassifier(n_estimators=100)
                    classifier.fit(X_train, y_train)

                    # Predict the classes
                    y_pred = classifier.predict(X_test)
                    y_prob = classifier.predict_proba(X_test)

                    # Append to the results
                    results["time"].append(i * segment_interval)
                    results["accuracy"].append(metrics.accuracy_score(y_test, y_pred))

                    for j, user in enumerate(y_test):
                        try:
                            uai = list(y_test).index(user)
                            uci = list(classifier.classes_).index(y_pred[j])
                            print(j, uci)
                        except ValueError:
                            continue
                        results["user_accuracies"][user].append(
                            1 if y_test[uai] == y_pred[uai] else 0
                        )
                        results["user_confidences"][user].append(y_prob[j][uci])

                    i += time_step
                    if i > 10000:
                        print("CHECK STOPPING CONDITION, LOOP NOT TERMINATING.")
                        break
                final_results["task"].append(task)
                final_results["window"].append(window)
                final_results["time"].append(results["time"])
                final_results["accuracy"].append(results["accuracy"])
                final_results["user_accuracies"].append(results["user_accuracies"])
                final_results["user_confidences"].append(results["user_confidences"])

        df = DataFrame(final_results)
        file_name = settings.RESULTS_DIR + settings.RESULTS_TEMPLATE.format(
            datetime.now().strftime("%Y%m%d_%H%M%S"), self.name
        )
        # Create the results dir if it does not exist
        if not isdir(settings.RESULTS_DIR):
            mkdir(settings.RESULTS_DIR)

        df.to_csv(
            file_name, index=False,
        )
        return file_name

    @staticmethod
    def _get_users_data(data):
        """
        Get seance id for each user in a form of a dict.
        """
        users = {}
        for user in list(set(data["user"])):
            x = data[data["user"] == user]
            users.update({user: sorted(list(set(x["seance"])))})
        return users
