from sklearn.tree import DecisionTreeClassifier

import settings
from helpers.helper import Helper

from plotly import graph_objs as go


class InfGainAnalysis(Helper):
    def __init__(self, datasets, criterion="entropy", interval=1, threshold=0.01):
        super().__init__()

        self.datasets = datasets
        self.criterion = criterion
        if self.criterion == "gini":
            self.criterion_name = "gini index"
        elif self.criterion == "entropy":
            self.criterion_name = "information gain"
        self.interval = interval
        self.threshold = threshold

        self.importance = None
        self.features = None

        self.gnuplot_script = "inf_gain.plt"

    def generate(self, task, show=False):
        """
        Use information gain and gini index to asses feature informativeness.
        """
        data = self.datasets[
            "{}_{}".format(str(task).zfill(2), str(self.interval).zfill(3))
        ]
        features = data.iloc[:, 2:]

        X = features.values
        y = ["User " + str(x) for x in data.iloc[:, 0].values]
        self.features = features.columns

        # define the model
        model = DecisionTreeClassifier(criterion=self.criterion)

        # fit the model
        model.fit(X, y)

        # get importance
        self.importance = model.feature_importances_

        if show:
            self._visualize(task)

    def _visualize(self, task):
        if self.importance is None:
            print("Call generate function first to calculate feature importance")
            return

        # Subset to only contain features with a score higher than the threshold
        results = {"features": [], "importance": []}
        for i, v in enumerate(self.importance):
            if abs(v) > self.threshold:
                results["features"].append(self.features[i])
                results["importance"].append(v)
        fig = go.Figure(data=[go.Bar(x=results["features"], y=results["importance"])])
        fig.update_layout(
            title="Feature importance calculated with {} for task {}".format(
                self.criterion_name, task
            ),
            xaxis_title="feature",
            yaxis_title="importance",
        )
        fig.show()

        with open(
            settings.VISUALIZATION_DIR + "inf_gain_{}.csv".format(str(task).zfill(2)),
            "w",
        ) as csv:
            csv.write("Feature\tValue\n")
            for i in range(0, len(results["features"])):
                csv.write(
                    "{}\t{}\n".format(results["features"][i], results["importance"][i])
                )
