import warnings
from datetime import datetime
from os import mkdir
from os.path import isdir
import random

from pandas import DataFrame, unique
from sklearn import metrics
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from plotly import graph_objs as go

import settings
from helpers.helper import Helper


class LdaAnalysis(Helper):
    def __init__(self, datasets, task, interval, users):
        super().__init__()
        self.datasets = datasets
        self.task = task
        self.interval = interval
        self.users = users
        self.gnuplot_script = "lda_plot.plt"

        warnings.filterwarnings("ignore")

    def generate(self, visualize=False):
        for user_number in self.users:
            all_users = list(
                unique(self.datasets["{}_128".format(str(self.task).zfill(2))]["user"])
            )
            users = random.sample(all_users, user_number)
            if user_number == 3:
                users = [10, 19, 21]
            elif user_number == 7:
                users = [7, 8, 10, 16, 18, 22, 26]
            elif user_number == 11:
                users = [7, 9, 13, 15, 16, 17, 18, 19, 23, 24, 27]
            data = self.datasets[
                "{}_{}".format(str(self.task).zfill(2), str(self.interval).zfill(3))
            ]

            visualization = []
            lda, score = self._get_lda(
                data, users, stats=False, score=True, single=True
            )
            if visualize:
                self._visualize(users, lda)
            for user in users:
                sub_data = lda[lda["user"] == user]
                visualization.append(sub_data["component_1"].values)
                visualization.append(sub_data["component_2"].values)
            self._to_file(
                "lda_data_{}_user_single".format(str(user_number).zfill(2)),
                visualization,
            )

            lda, score = self._get_lda(data, users, stats=False, score=True)
            visualization = []
            if visualize:
                self._visualize(users, lda, single=False)
            for user in users:
                sub_data = lda[lda["user"] == user]
                visualization.append(sub_data["component_1"].values)
                visualization.append(sub_data["component_2"].values)
            self._to_file(
                "lda_data_{}_user_double".format(str(user_number).zfill(2)),
                visualization,
            )

    @staticmethod
    def _visualize(users, lda, single=True):
        graphs = []
        for user in users:
            sub_data = lda[lda["user"] == user]
            graphs.append(
                go.Scatter(
                    x=sub_data["component_1"],
                    y=sub_data["component_2"],
                    name="User {}".format(user),
                    mode="markers",
                )
            )
        fig = go.Figure(graphs)
        if single:
            title = "Single seance per user"
        else:
            title = "Multiple seances per user"
        fig.update_layout(
            title=title,
            xaxis_title="Component 1",
            yaxis_title="Component 2",
            showlegend=True,
        )
        fig.show()

    @staticmethod
    def _get_lda(
        data, users, comp_num=2, stats=True, score=False, by_user=False, single=False
    ):
        if single:
            seances = []
            for user in users:
                seances.append(data[data["user"] == user].iloc[0, 1])
            data = data[data.seance.isin(seances)]
        else:
            data = data[data.user.isin(users)]

        X = data.iloc[:, 2:].values
        y = data.iloc[:, 0].values.ravel()
        z = data.iloc[:, 1].values.ravel()

        lda = LinearDiscriminantAnalysis(n_components=comp_num)
        if not by_user:
            X_lda = lda.fit_transform(X, z)
            score = round(metrics.calinski_harabasz_score(X_lda, z))
        else:
            X_lda = lda.fit_transform(X, y)
            score = round(metrics.calinski_harabasz_score(X_lda, y))
        if stats:
            print(lda.explained_variance_ratio_)
            print(score)

        y = y.reshape(len(y), 1)
        z = z.reshape(len(z), 1)
        df = DataFrame(
            [list(y) + list(z) + list(x) for x, y, z in zip(X_lda, y, z)],
            columns=["user", "try", "component_1", "component_2"],
        )
        if score:
            return df, score
        return df

    @staticmethod
    def _to_file(file_name, visualization):
        # Create the visualization dir if it does not exist
        if not isdir(settings.VISUALIZATION_DIR):
            mkdir(settings.VISUALIZATION_DIR)

        file_name = settings.VISUALIZATION_DIR + "/" + file_name + ".csv"

        with open(file_name, "w") as file:
            i = 0
            stop = False
            while not stop:
                row = ""
                stop = True
                for col in visualization:
                    if len(col) > i:
                        row += str(col[i]) + "\t"
                        stop = False
                    else:
                        row += "\t"
                file.write(row + "\n")
                i += 1
