import warnings
from datetime import timedelta
from os import mkdir
from os.path import isdir

from pandas import read_csv, unique
from plotly import graph_objects as go

from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.dummy import DummyClassifier
from sklearn.ensemble import (
    AdaBoostClassifier,
    RandomForestClassifier,
)
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.tree import DecisionTreeClassifier

import settings
from helpers.helper import Helper
from identification.results_generator import ResultsGenerator


class MlComparison(Helper):
    def __init__(self, datasets=None, results=None):
        super().__init__()
        self.datasets = datasets
        self.results = results

        self.name = "ml_comparison"
        self.gnuplot_script = "model_comparison.plt"

        self.segment_intervals = [1, 2, 4, 8, 16, 32, 64, 128]
        self.sensor_combinations = [
            ["acc"],
            ["gyr"],
            ["frc"],
            ["cpu", "mem", "net"],
            ["acc", "gyr"],
            ["acc", "frc"],
            ["gyr", "frc"],
            ["acc", "gyr", "frc"],
            ["acc", "gyr", "frc", "cpu", "mem", "net"],
            ["acc", "gyr", "frc", "cpu", "mem", "net", "pir"],
        ]

        warnings.filterwarnings("ignore")

    def generate(self, cache=True):
        """
        Generate the results for multiple ml algorithms.
        """
        if cache:
            self.results = self._check_for_results()
            if self.results:
                print("Using cached results.")
                return
        random_grid_forest = {
            "bootstrap": True,
            "max_depth": 60,
            "max_features": "auto",
            "min_samples_leaf": 2,
            "min_samples_split": 5,
            "n_estimators": 800,
        }
        grid_ada = {
            "n_estimators": 1000,
            "base_estimator": DecisionTreeClassifier(
                max_depth=20,
                min_samples_leaf=4,
                min_samples_split=10,
                max_features="auto",
            ),
        }

        approach = ResultsGenerator(
            datasets=self.datasets,
            epochs=1,
            tasks=settings.TASKS,
            number_of_users=[-1],
            sensor_combinations=[["acc", "gyr", "frc", "cpu", "mem", "net", "pir"]],
            segment_intervals=[1, 2, 4, 8, 16, 32, 64],
            # segment_intervals=[16],
            time_limit=timedelta(seconds=180),
            models=[
                (DummyClassifier, {"strategy": "most_frequent"}, "MC"),
                # (SVC, {"gamma": "auto", "probability": True}, "SVM"),
                (KNeighborsClassifier, {"n_neighbors": 3}, "KNN"),
                (GaussianNB, {"var_smoothing": 1e-9}, "NB"),
                (LinearDiscriminantAnalysis, {"n_components": 2}, "LDA"),
                # (GradientBoostingClassifier, grid_gradient, "GB"),
                (AdaBoostClassifier, grid_ada, "ADA"),
                (RandomForestClassifier, random_grid_forest, "RF"),
            ],
            name=self.name,
        )
        self.results = approach.generate()

    def visualize(self, show=True):
        """
        Visualize the resulting csv file and generate a gnuplot input csv file.
        """
        df = read_csv(self.results)
        TASKS = list(unique(df["task"]))

        graphs = []
        results = {"task": [], "model": [], "accuracy": []}
        for task in TASKS:
            sub = df[df["task"] == task]
            results["task"].extend(sub["task"])
            results["model"].extend(sub["model"])
            results["accuracy"].extend(sub["accuracy"])
            graphs.append(
                go.Bar(x=sub["model"], y=sub["accuracy"], name="Task {}".format(task))
            )
        title = "Comparison of the accuracy of the ml algorithms "
        fig = go.Figure(data=graphs)
        fig.update_layout(
            title=title,
            xaxis_title="Model",
            xaxis_type="category",
            yaxis_title="Accuracy",
            yaxis_range=[0, 1],
            barmode="group",
        )


        df = read_csv(self.results)
        TASKS = list(unique(df["task"]))

        for interval in self.segment_intervals:
            for task in TASKS:
                graphs = []
                results = {"sensors": [], "model": [], "accuracy": []}
                for sensor in self.sensor_combinations:

                    sub = df[
                        (df["task"] == task)
                        & (df["segment_interval"] == interval)
                        & (df["sensors"] == "".join(sorted([x[0] for x in sensor])))
                    ]
                    results["sensors"].extend(sub["sensors"])
                    results["model"].extend(sub["model"])
                    results["accuracy"].extend(sub["accuracy"])
                    graphs.append(
                        go.Bar(
                            x=sub["model"], y=sub["accuracy"], name="{}".format(sensor)
                        )
                    )
                title = (
                    "Comparison of the accuracy of the ml algorithms. "
                    "Task {}. Segment interval of {} seconds.".format(task, interval)
                )
                fig = go.Figure(data=graphs)
                fig.update_layout(
                    title=title,
                    xaxis_title="Model",
                    xaxis_type="category",
                    yaxis_title="Accuracy",
                    yaxis_range=[0, 1],
                    barmode="group",
                )
                if show:
                    fig.show()
        print(df[["task", "model", "accuracy"]])

        # Create the visualization dir if it does not exist
        if not isdir(settings.VISUALIZATION_DIR):
            mkdir(settings.VISUALIZATION_DIR)

        with open(settings.VISUALIZATION_DIR + "model_comparison.csv", "w") as csv:
            csv.write("Model\t")
            models = list(unique(df["model"]))
            for model in models:
                csv.write(model + "\t")
            csv.write("\n")
            for task in sorted(list(unique(df["task"]))):
                csv.write(str(task) + "\t")
                for model in models:
                    acc = str(
                        df[(df["task"] == task) & (df["model"] == model)][
                            "accuracy"
                        ].values[0]
                    )
                    csv.write(acc + "\t")
                csv.write("\n")
