import random
from datetime import datetime
from os import mkdir
from os.path import isdir

from IPython.core.display import clear_output
from numpy import argsort, where, mean, array, nanmean, std, nanstd
from pandas import unique, DataFrame
from sklearn.feature_selection import SelectKBest, mutual_info_classif
from sklearn.metrics import accuracy_score
from sklearn.model_selection import GridSearchCV

import settings


class ResultsGenerator:
    def __init__(
        self,
        datasets,
        epochs,
        tasks,
        number_of_users,
        segment_intervals,
        time_limit,
        models,
        name,
    ):
        """
        :param datasets: dict of datasets with keys x_y where x (2 zfill digits) represents
                         a task and y (3 zfill digits) time segment intervals
        """
        self.datasets = datasets
        self.epochs = epochs
        self.tasks = tasks
        self.number_of_users = number_of_users
        self.segment_intervals = segment_intervals
        self.time_limit = time_limit
        self.models = models
        self.name = name

    def generate(self):
        """
        Generate the results of the given result generating function.
        """
        num_of_runs = (
            len(self.tasks) * len(self.segment_intervals) * len(self.number_of_users)
        )
        per_task_limit = self.time_limit / num_of_runs
        run = 0
        results = {
            "model": [],
            "task": [],
            "segment_interval": [],
            "number_of_users": [],
            "accuracy": [],
            "accuracy_sd": [],
            "true_confidence": [],
            "true_confidence_sd": [],
            "false_confidence": [],
            "false_confidence_sd": [],
            "user_accuracies": [],
            "user_accuracies_sd": [],
            "user_true_confidences": [],
            "user_true_confidences_sd": [],
            "user_false_confidences": [],
            "user_false_confidences_sd": [],
            "feature_frequency": [],
            "epochs": [],
        }
        for number_of_users in self.number_of_users:
            for segment_interval in self.segment_intervals:
                for task in self.tasks:
                    run += 1

                    # Get the data according to the current task and segment interval
                    try:
                        data = self.datasets[
                            "{}_{}".format(
                                str(task).zfill(2), str(segment_interval).zfill(3)
                            )
                        ]
                    except KeyError:
                        continue

                    # Get all the users present in the current dataset
                    all_users = list(unique(data["user"]))

                    # Temporary, result holding lists
                    models = [x[2] for x in self.models]
                    accuracies = {}
                    confidences = {}
                    u_accuracies = {}
                    u_confidences = {}
                    feature_sets = {}
                    epochs = {}
                    for model in models:
                        accuracies.update({model: []})
                        confidences.update({model: []})
                        u_accuracies.update({model: []})
                        u_confidences.update({model: []})
                        feature_sets.update({model: []})
                        epochs.update({model: []})

                    s = datetime.now()
                    for e in range(self.epochs):
                        # Report on progress
                        clear_output()
                        print("Run: {} of {}".format(run, num_of_runs))
                        print("Epoch: {} of {}".format(e + 1, self.epochs))
                        print(
                            "Terminating in: {}".format(
                                per_task_limit - (datetime.now() - s)
                            )
                        )
                        print()
                        print(
                            "ETA: {}".format(
                                (num_of_runs - run) * per_task_limit
                                + per_task_limit
                                - (datetime.now() - s)
                            )
                        )
                        print()

                        # Generate the data of one epoch
                        epoch_results = self._use_ml(data, number_of_users)

                        # Save this epochs' data
                        for model in models:
                            accuracies[model].append(epoch_results[model]["accuracy"])
                            confidences[model].append(
                                epoch_results[model]["confidence"]
                            )
                            u_accuracies[model].append(
                                epoch_results[model]["user_accuracies"]
                            )
                            u_confidences[model].append(
                                epoch_results[model]["user_confidences"]
                            )
                            feature_sets[model].append(
                                epoch_results[model]["feature_set"]
                            )

                        # Break the loop if per_task_limit has been reached
                        if (
                            datetime.now() - s
                        ).total_seconds() > per_task_limit.total_seconds():
                            print("Time limit reached!")
                            print("Moving on to the next task...")
                            break

                    # Extract the results for the last parameter run and save them in the final dict
                    for model in models:
                        user_accuracies = {}
                        user_accuracies_sd = {}
                        user_confidences = {}
                        for user in all_users:
                            user_accuracies.update({user: []})
                            user_accuracies_sd.update({user: []})
                            user_confidences.update({user: {"true": [], "false": []}})
                        # Calculate mean per user accuracy
                        a = array(u_accuracies[model])
                        for y in list(
                            [
                                (x[0], x[1])
                                for x in a.reshape(a.shape[0] * a.shape[1], -1)
                            ]
                        ):
                            user_accuracies[y[0]].append(y[1])
                        for user in user_accuracies:
                            user_accuracies_sd[user] = std(user_accuracies[user])
                            user_accuracies[user] = mean(user_accuracies[user])

                        # Calculate mean per user true and false confidence rate
                        for x in u_confidences[model]:
                            for user in x:
                                user_confidences[user]["true"].append(x[user]["true"])
                                user_confidences[user]["false"].append(x[user]["false"])

                        true_user_confidences = {}
                        false_user_confidences = {}
                        true_user_confidences_sd = {}
                        false_user_confidences_sd = {}
                        for user in user_confidences:
                            true_user_confidences.update(
                                {user: nanmean(user_confidences[user]["true"])}
                            )
                            false_user_confidences.update(
                                {user: nanmean(user_confidences[user]["false"])}
                            )
                            true_user_confidences_sd.update(
                                {user: nanstd(user_confidences[user]["true"])}
                            )
                            false_user_confidences_sd.update(
                                {user: nanstd(user_confidences[user]["false"])}
                            )

                        # Calculate each feature frequency
                        feature_frequencies = {}
                        for feature_set in feature_sets[model]:
                            for feature in feature_set:
                                if feature in feature_frequencies:
                                    feature_frequencies[feature] += 1
                                else:
                                    feature_frequencies.update({feature: 1})
                        for feature in feature_frequencies:
                            feature_frequencies[feature] = feature_frequencies[
                                feature
                            ] / (e + 1)
                        results["model"].append(model)
                        results["task"].append(task)
                        results["segment_interval"].append(segment_interval)
                        results["number_of_users"].append(number_of_users)
                        results["accuracy"].append(mean(accuracies[model]))
                        results["accuracy_sd"].append(std(accuracies[model]))
                        results["true_confidence"].append(
                            nanmean([x["true"] for x in confidences[model]])
                        )
                        results["true_confidence_sd"].append(
                            nanstd([x["true"] for x in confidences[model]])
                        )
                        results["false_confidence"].append(
                            nanmean([x["false"] for x in confidences[model]])
                        )
                        results["false_confidence_sd"].append(
                            nanstd([x["false"] for x in confidences[model]])
                        )
                        results["user_accuracies"].append(user_accuracies)
                        results["user_accuracies_sd"].append(user_accuracies_sd)
                        results["user_true_confidences"].append(true_user_confidences)
                        results["user_true_confidences_sd"].append(
                            true_user_confidences_sd
                        )
                        results["user_false_confidences"].append(false_user_confidences)
                        results["user_false_confidences_sd"].append(
                            false_user_confidences_sd
                        )
                        results["feature_frequency"].append(feature_frequencies)
                        results["epochs"].append(e + 1)

        df = DataFrame(results)
        file_name = settings.RESULTS_DIR + settings.RESULTS_TEMPLATE.format(
            datetime.now().strftime("%Y%m%d_%H%M%S"), self.name
        )
        # Create the results dir if it does not exist
        if not isdir(settings.RESULTS_DIR):
            mkdir(settings.RESULTS_DIR)

        df.to_csv(
            file_name, index=False,
        )
        return file_name

    def _use_ml(self, data, number_of_users):
        """
        Take self.number_of_users random users and return the results.
        """
        # Take n random users
        all_users = list(unique(data["user"]))
        if number_of_users != -1:
            users = sorted(random.sample(all_users, number_of_users))
        else:
            users = all_users
            number_of_users = len(users)
        sub_data = data[data.user.isin(users)]

        # Split the data
        seances = self._get_users_data(sub_data)
        train_seances = [x[0] for x in seances.values()]
        test_seances = [x[1] for x in seances.values()]
        training_set = sub_data[sub_data["seance"].isin(train_seances)]
        testing_set = sub_data[sub_data["seance"].isin(test_seances)]
        X = training_set.iloc[:, 2:].values
        y = training_set.iloc[:, 0].values.ravel()

        # Feature selection
        cols = list(sub_data.columns)
        f_cols = list(sub_data.iloc[:, 2:].columns)
        sel = SelectKBest(mutual_info_classif, k=2)
        sel.fit(X, y)
        indices = argsort(sel.scores_)
        feature_set = cols[0:2] + [f_cols[x] for x in indices[-10:]]

        # Take the selected features
        X_train = training_set[feature_set].iloc[:, 2:].values
        y_train = training_set[feature_set].iloc[:, 0].values.ravel()
        X_test = testing_set[feature_set].iloc[:, 2:].values
        y_test = testing_set[feature_set].iloc[:, 0].values.ravel()

        # Final results
        results = {}

        for i, model in enumerate(self.models):
            model, params, model_name = model
            if params:
                search = False
                for param in params.values():
                    if isinstance(param, list) and len(param) > 1:
                        search = True
                if search:
                    classifier = GridSearchCV(
                        estimator=model(), param_grid=params, cv=3, n_jobs=8, verbose=0,
                    )
                else:
                    # Grid search
                    classifier = model()
                    classifier.set_params(**params)
                # Some models do not need parameter space search.
            else:
                classifier = model()

            classifier.fit(X_train, y_train)

            # Get accuracy
            y_pred = classifier.predict(X_test)
            accuracy = accuracy_score(y_test, y_pred)

            # Calculate per user accuracy
            user_accuracies = []
            for user in users:
                idx = where(y_test == user)
                test = y_test[idx]
                pred = y_pred[idx]
                score = 0
                for i, _ in enumerate(test):
                    if test[i] == pred[i]:
                        score += 1
                user_accuracies.append((user, score / len(test)))

            # Initialize confidence levels results dicts
            user_confidences = {}
            for user in users:
                user_confidences.update({user: {"true": [], "false": []}})

            # Get per class probabilities
            y_prob = classifier.predict_proba(X_test)

            # Take the probability of the classified user for every test
            # case and append it to the appropriate results list
            for i, true_class in enumerate(y_test):
                conf = y_prob[i][users.index(y_pred[i])]
                if true_class == y_pred[i]:
                    key = "true"
                else:
                    key = "false"
                user_confidences[true_class][key].append(conf)

            for key in ["true", "false"]:
                for user in users:
                    user_confidences[user][key] = mean(user_confidences[user][key])

            # Calculate the overall true and false mean confidence levels
            true_conf = []
            false_conf = []
            for user in users:
                true_conf.append(user_confidences[user]["true"])
                false_conf.append(user_confidences[user]["false"])

            confidence = {"true": nanmean(true_conf), "false": nanmean(false_conf)}

            # Append to results
            results.update(
                {
                    model_name: {
                        "accuracy": accuracy,
                        "confidence": confidence,
                        "user_accuracies": user_accuracies,
                        "user_confidences": user_confidences,
                        "feature_set": feature_set[2:],
                    }
                }
            )

        return results

    @staticmethod
    def _get_users_data(data):
        """
        Get seance id for each user in a form of a dict.
        """
        users = {}
        for user in list(set(data["user"])):
            x = data[data["user"] == user]
            users.update({user: sorted(list(set(x["seance"])))})
        return users
