import subprocess
import warnings
from datetime import timedelta

from numpy import mean, std
from pandas import read_csv, unique, DataFrame
from sklearn.ensemble import RandomForestClassifier

import settings
from helpers.helper import Helper
from identification.results_generator_backup import ResultsGenerator
from plotly import graph_objects as go


class UserIntervals(Helper):
    def __init__(self, datasets=None, results=None):
        super().__init__()
        self.datasets = datasets
        self.results = results

        self.name = "users_intervals"
        self.gnuplot_script = "users_vs_intervals.plt"

        warnings.filterwarnings("ignore")

    def generate(self, cache=True):
        """
        Generate the results for multiple ml algorithms.
        """
        if cache:
            self.results = self._check_for_results()
            if self.results:
                print("Using cached results.")
                return
        random_grid_forest = {
            "bootstrap": True,
            "max_depth": 60,
            "max_features": "auto",
            "min_samples_leaf": 2,
            "min_samples_split": 5,
            "n_estimators": 800,
        }

        approach = ResultsGenerator(
            self.datasets,
            500,
            settings.TASKS,
            [3, 7, 11, 15],
            [1, 4, 8, 32, 128],
            timedelta(hours=24),
            [(RandomForestClassifier, random_grid_forest, "RF")],
            "users_intervals",
        )
        self.results = approach.generate()

    def visualize(self, show=True):
        """
        Visualize the resulting csv file and generate a gnuplot input csv file.
        """
        df = read_csv(self.results)
        NUMBER_OF_USERS = list(unique(df["number_of_users"]))
        SEGMENT_INTERVALS = list(unique(df["segment_interval"]))

        graphs = []

        results = {"number_of_users": [], "segment_interval": [], "avg": [], "std": []}

        for user in NUMBER_OF_USERS:
            sub = df[df["number_of_users"] == user]
            accuracies = []
            stds = []
            for interval in SEGMENT_INTERVALS:
                avg = mean(sub[sub["segment_interval"] == interval]["accuracy"])
                sd = std(sub[sub["segment_interval"] == interval]["accuracy"])
                accuracies.append(avg)
                stds.append(sd)
                results["number_of_users"].append(user)
                results["segment_interval"].append(interval)
                results["avg"].append(avg)
                results["std"].append(sd)

            graphs.append(
                go.Bar(
                    x=SEGMENT_INTERVALS,
                    y=accuracies,
                    name="{} users".format(user),
                    error_y={"type": "data", "array": stds, "visible": True,},
                )
            )

        title = "Number of users vs. classification accuracy."

        fig = go.Figure(data=graphs)
        fig.update_layout(
            title=title,
            xaxis_title="Segment interval",
            xaxis_type="category",
            yaxis_title="Accuracy",
            yaxis_range=[0, 1],
            barmode="group",
        )
        if show:
            fig.show()

        df = DataFrame(results)
        with open(settings.VISUALIZATION_DIR + "users_vs_intervals.csv", "w") as csv:
            csv.write("User\t")
            for user in NUMBER_OF_USERS:
                csv.write(str(user) + "\t")
            for user in NUMBER_OF_USERS:
                csv.write(str(user) + "\t")
            csv.write("\n")
            for interval in SEGMENT_INTERVALS:
                csv.write(str(interval) + "\t")
                for user in NUMBER_OF_USERS:
                    acc = str(
                        df[
                            (df["number_of_users"] == user)
                            & (df["segment_interval"] == interval)
                        ]["avg"].values[0]
                    )
                    csv.write(acc + "\t")
                for user in NUMBER_OF_USERS:
                    sd = str(
                        df[
                            (df["number_of_users"] == user)
                            & (df["segment_interval"] == interval)
                        ]["std"].values[0]
                    )
                    csv.write(sd + "\t")
                csv.write("\n")
