from datetime import timedelta, datetime
from math import sqrt

from IPython.core.display import clear_output
import numpy as np
from numpy import std, mean, argsort, array
from numpy.fft import fft, fftfreq
from pandas import DataFrame, read_csv
from scipy.interpolate import interp1d
from scipy.signal import find_peaks, get_window

import settings


class CsvDataHandler:
    def __init__(self, data=None, task=None, interval=None):
        self.data = data
        self.task = task
        self.interval = interval

    # GENERATING CSV DATA
    def generate_csv_data(self):
        """
        Generate csv file with calculated features, from data subsampled to the given time interval.
        """
        clear_output()
        step = timedelta(seconds=self.interval)

        print(
            "Generating segmented data csv file with {} seconds intervals (task {}) for {} seances.".format(
                self.interval, self.task, len(self.data.seances)
            )
        )

        file_name = settings.DATA_DIR + settings.CSV_TEMPLATE.format(
            self.task, self.interval
        )
        resulting_data = {
            "user": [],
            "seance": [],
            "ax_me": [],
            "ax_sd": [],
            "ax_mcr": [],
            "ax_mai": [],
            "ay_me": [],
            "ay_sd": [],
            "ay_mcr": [],
            "ay_mai": [],
            "az_me": [],
            "az_sd": [],
            "az_mcr": [],
            "az_mai": [],
            "a_me": [],
            "a_sd": [],
            "a_mcr": [],
            "a_mai": [],
            "gx_me": [],
            "gx_sd": [],
            "gx_mcr": [],
            "gy_me": [],
            "gy_sd": [],
            "gy_mcr": [],
            "gz_me": [],
            "gz_sd": [],
            "gz_mcr": [],
            "g_me": [],
            "g_sd": [],
            "g_mcr": [],
            "fa_me": [],
            "fa_sd": [],
            "fa_mcr": [],
            "fb_me": [],
            "fb_sd": [],
            "fb_mcr": [],
            "fc_me": [],
            "fc_sd": [],
            "fc_mcr": [],
            "fd_me": [],
            "fd_sd": [],
            "fd_mcr": [],
            "ca_me": [],
            "ca_sd": [],
            "ca_min": [],
            "ca_max": [],
            "ca_mcr": [],
            "cb_me": [],
            "cb_sd": [],
            "cb_min": [],
            "cb_max": [],
            "cb_mcr": [],
            "cc_me": [],
            "cc_sd": [],
            "cc_min": [],
            "cc_max": [],
            "cc_mcr": [],
            "cd_me": [],
            "cd_sd": [],
            "cd_min": [],
            "cd_max": [],
            "cd_mcr": [],
            "c_me": [],
            "c_sd": [],
            "c_min": [],
            "c_max": [],
            "c_mcr": [],
            "m_me": [],
            "m_sd": [],
            "m_min": [],
            "m_max": [],
            "m_jc": [],
            "m_jr": [],
            "m_jv": [],
            "m_iji": [],
            "ns_me": [],
            "ns_sd": [],
            "ns_sum": [],
            "ns_jc": [],
            "ns_jr": [],
            "ns_jv": [],
            "ns_iji": [],
            "nr_me": [],
            "nr_sd": [],
            "nr_sum": [],
            "nr_jc": [],
            "nr_jr": [],
            "nr_jv": [],
            "nr_iji": [],
            "ir_me": [],
            "ir_sd": [],
            "ax_bin": [],
            "ay_bin": [],
            "az_bin": [],
            "gx_bin": [],
            "gy_bin": [],
            "gz_bin": [],
            "fa_bin": [],
            "fb_bin": [],
            "fc_bin": [],
            "fd_bin": [],
        }

        count = 1
        for seance_id in self.data.seances:
            seance = self.data.get_data_for_seance(seance_id)
            print(
                "-----------------------------------------------------------------------------"
            )
            print(
                "{} of {}, seance {}".format(count, len(self.data.seances), seance_id)
            )
            count += 1
            start = datetime.strptime(
                self.data.start_ends[seance_id][0], "%Y-%m-%d %H:%M:%S.%f+02"
            )
            end = datetime.strptime(
                self.data.start_ends[seance_id][1], "%Y-%m-%d %H:%M:%S.%f+02"
            )

            data = {}
            for sensor in self.data.sensors:
                sub_df = seance[seance["sensor"] == sensor].sort_values(by="timestamp")[
                    ["value", "timestamp"]
                ]
                data.update({sensor: sub_df})

            i = 1
            while start < end:
                sub_data = {}
                for sensor in data:
                    sub_df = data[sensor]
                    sub_df = sub_df[
                        (sub_df["timestamp"] >= start)
                        & (sub_df["timestamp"] < start + step)
                    ]
                    sub_data.update({sensor: sub_df})

                # accelerometer
                ax_val, ax_me, ax_sd = self._process_signal(
                    list(sub_data["accel01_x"]["value"])
                )
                ay_val, ay_me, ay_sd = self._process_signal(
                    list(sub_data["accel01_y"]["value"])
                )
                az_val, az_me, az_sd = self._process_signal(
                    list(sub_data["accel01_z"]["value"])
                )
                a_val, a_me, a_sd = self._join_accelerometer_signals(
                    ax_val, ay_val, az_val
                )
                ax_mcr = self._mean_crossing_rate(ax_val, ax_me)
                ay_mcr = self._mean_crossing_rate(ay_val, ay_me)
                az_mcr = self._mean_crossing_rate(az_val, az_me)
                a_mcr = self._mean_crossing_rate(a_val, a_me)
                ax_mai = self._mean_acceleration_intensity(ax_val)
                ay_mai = self._mean_acceleration_intensity(ay_val)
                az_mai = self._mean_acceleration_intensity(az_val)
                a_mai = self._mean_acceleration_intensity(a_val)

                ax_bin = self._get_10_mag_bins(
                    list(sub_data["accel01_x"]["value"]),
                    list(sub_data["accel01_x"]["timestamp"]),
                )
                ay_bin = self._get_10_mag_bins(
                    list(sub_data["accel01_y"]["value"]),
                    list(sub_data["accel01_y"]["timestamp"]),
                )
                az_bin = self._get_10_mag_bins(
                    list(sub_data["accel01_z"]["value"]),
                    list(sub_data["accel01_z"]["timestamp"]),
                )

                # gyroscope
                gx_val, gx_me, gx_sd = self._process_signal(
                    list(sub_data["gyro01_x"]["value"])
                )
                gy_val, gy_me, gy_sd = self._process_signal(
                    list(sub_data["gyro01_y"]["value"])
                )
                gz_val, gz_me, gz_sd = self._process_signal(
                    list(sub_data["gyro01_z"]["value"])
                )
                g_val, g_me, g_sd = self._join_accelerometer_signals(
                    gx_val, gy_val, gz_val
                )
                gx_mcr = self._mean_crossing_rate(gx_val, gx_me)
                gy_mcr = self._mean_crossing_rate(gy_val, gy_me)
                gz_mcr = self._mean_crossing_rate(gz_val, gz_me)
                g_mcr = self._mean_crossing_rate(g_val, g_me)
                gx_bin = self._get_10_mag_bins(
                    list(sub_data["gyro01_x"]["value"]),
                    list(sub_data["gyro01_x"]["timestamp"]),
                )
                gy_bin = self._get_10_mag_bins(
                    list(sub_data["gyro01_y"]["value"]),
                    list(sub_data["gyro01_y"]["timestamp"]),
                )
                gz_bin = self._get_10_mag_bins(
                    list(sub_data["gyro01_z"]["value"]),
                    list(sub_data["gyro01_z"]["timestamp"]),
                )

                # force
                fa_val, fa_me, fa_sd = self._process_signal(
                    list(sub_data["fsr_01"]["value"])
                )
                fb_val, fb_me, fb_sd = self._process_signal(
                    list(sub_data["fsr_02"]["value"])
                )
                fc_val, fc_me, fc_sd = self._process_signal(
                    list(sub_data["fsr_03"]["value"])
                )
                fd_val, fd_me, fd_sd = self._process_signal(
                    list(sub_data["fsr_04"]["value"])
                )
                fa_mcr = self._mean_crossing_rate(fa_val, fa_me)
                fb_mcr = self._mean_crossing_rate(fb_val, fb_me)
                fc_mcr = self._mean_crossing_rate(fc_val, fc_me)
                fd_mcr = self._mean_crossing_rate(fd_val, fd_me)
                fa_bin = self._get_10_mag_bins(
                    list(sub_data["fsr_01"]["value"]),
                    list(sub_data["fsr_01"]["timestamp"]),
                )
                fb_bin = self._get_10_mag_bins(
                    list(sub_data["fsr_02"]["value"]),
                    list(sub_data["fsr_02"]["timestamp"]),
                )
                fc_bin = self._get_10_mag_bins(
                    list(sub_data["fsr_03"]["value"]),
                    list(sub_data["fsr_03"]["timestamp"]),
                )
                fd_bin = self._get_10_mag_bins(
                    list(sub_data["fsr_04"]["value"]),
                    list(sub_data["fsr_04"]["timestamp"]),
                )

                # cpu
                print(sub_data["cpuusage_01"]["value"])
                ca_val, ca_me, ca_sd = self._process_signal(
                    list(sub_data["cpuusage_01"]["value"])
                )
                print(ca_val)
                print(ca_me)
                print(ca_sd)
                cb_val, cb_me, cb_sd = self._process_signal(
                    list(sub_data["cpuusage_02"]["value"])
                )
                cc_val, cc_me, cc_sd = self._process_signal(
                    list(sub_data["cpuusage_03"]["value"])
                )
                cd_val, cd_me, cd_sd = self._process_signal(
                    list(sub_data["cpuusage_04"]["value"])
                )
                c_val, c_me, c_sd = self._join_cpu_signals(
                    ca_val, cb_val, cc_val, cd_val
                )
                ca_min, ca_max, ca_mcr = self._get_cpu_stats(ca_val)
                cb_min, cb_max, cb_mcr = self._get_cpu_stats(cb_val)
                cc_min, cc_max, cc_mcr = self._get_cpu_stats(cc_val)
                cd_min, cd_max, cd_mcr = self._get_cpu_stats(cd_val)
                c_min, c_max, c_mcr = self._get_cpu_stats(c_val)

                # ram
                m_val, m_me, m_sd = self._process_signal(
                    list(sub_data["mempercentage_01"]["value"])
                )
                derivatives, peaks = self._find_ram_jump(m_val)
                m_me, m_min, m_max, m_jc, m_jr, m_jv, m_iji = self._get_mem_stats(
                    m_val, peaks, derivatives
                )

                # net
                ns_val, ns_me, ns_sd = self._process_signal(
                    list(sub_data["netpacketssent_01"]["value"])
                )
                nr_val, nr_me, nr_sd = self._process_signal(
                    list(sub_data["netpacketsreceived_01"]["value"])
                )
                ns_der, ns_pe = self._find_net_jump(ns_val)
                nr_der, nr_pe = self._find_net_jump(nr_val)
                ns_sum, ns_jc, ns_jr, ns_jv, ns_iji = self._get_net_stats(
                    ns_val, ns_pe, ns_der
                )
                nr_sum, nr_jc, nr_jr, nr_jv, nr_iji = self._get_net_stats(
                    nr_val, nr_pe, nr_der
                )

                # ir sensors
                pir_data = self._get_pir_data(sub_data)
                ir_me, ir_sd = self._process_pir_data(
                    pir_data, start, start + step, self.interval
                )

                resulting_data["user"].append(list(seance["user_id"])[0])
                resulting_data["seance"].append(list(seance["seance_id"])[0])
                resulting_data["ax_me"].append(ax_me)
                resulting_data["ax_sd"].append(ax_sd)
                resulting_data["ax_mcr"].append(ax_mcr)
                resulting_data["ax_mai"].append(ax_mai)
                resulting_data["ay_me"].append(ay_me)
                resulting_data["ay_sd"].append(ay_sd)
                resulting_data["ay_mcr"].append(ay_mcr)
                resulting_data["ay_mai"].append(ay_mai)
                resulting_data["az_me"].append(az_me)
                resulting_data["az_sd"].append(az_sd)
                resulting_data["az_mcr"].append(az_mcr)
                resulting_data["az_mai"].append(az_mai)
                resulting_data["a_me"].append(a_me)
                resulting_data["a_sd"].append(a_sd)
                resulting_data["a_mcr"].append(a_mcr)
                resulting_data["a_mai"].append(a_mai)
                resulting_data["gx_me"].append(gx_me)
                resulting_data["gx_sd"].append(gx_sd)
                resulting_data["gx_mcr"].append(gx_mcr)
                resulting_data["gy_me"].append(gy_me)
                resulting_data["gy_sd"].append(gy_sd)
                resulting_data["gy_mcr"].append(gy_mcr)
                resulting_data["gz_me"].append(gz_me)
                resulting_data["gz_sd"].append(gz_sd)
                resulting_data["gz_mcr"].append(gz_mcr)
                resulting_data["g_me"].append(g_me)
                resulting_data["g_sd"].append(g_sd)
                resulting_data["g_mcr"].append(g_mcr)
                resulting_data["fa_me"].append(fa_me)
                resulting_data["fa_sd"].append(fa_sd)
                resulting_data["fa_mcr"].append(fa_mcr)
                resulting_data["fb_me"].append(fb_me)
                resulting_data["fb_sd"].append(fb_sd)
                resulting_data["fb_mcr"].append(fb_mcr)
                resulting_data["fc_me"].append(fc_me)
                resulting_data["fc_sd"].append(fc_sd)
                resulting_data["fc_mcr"].append(fc_mcr)
                resulting_data["fd_me"].append(fd_me)
                resulting_data["fd_sd"].append(fd_sd)
                resulting_data["fd_mcr"].append(fd_mcr)
                resulting_data["ca_me"].append(ca_me)
                resulting_data["ca_sd"].append(ca_sd)
                resulting_data["ca_min"].append(ca_min)
                resulting_data["ca_max"].append(ca_max)
                resulting_data["ca_mcr"].append(ca_mcr)
                resulting_data["cb_me"].append(cb_me)
                resulting_data["cb_sd"].append(cb_sd)
                resulting_data["cb_min"].append(cb_min)
                resulting_data["cb_max"].append(cb_max)
                resulting_data["cb_mcr"].append(cb_mcr)
                resulting_data["cc_me"].append(cc_me)
                resulting_data["cc_sd"].append(cc_sd)
                resulting_data["cc_min"].append(cc_min)
                resulting_data["cc_max"].append(cc_max)
                resulting_data["cc_mcr"].append(cc_mcr)
                resulting_data["cd_me"].append(cd_me)
                resulting_data["cd_sd"].append(cd_sd)
                resulting_data["cd_min"].append(cd_min)
                resulting_data["cd_max"].append(cd_max)
                resulting_data["cd_mcr"].append(cd_mcr)
                resulting_data["c_me"].append(c_me)
                resulting_data["c_sd"].append(c_sd)
                resulting_data["c_min"].append(c_min)
                resulting_data["c_max"].append(c_max)
                resulting_data["c_mcr"].append(c_mcr)
                resulting_data["m_me"].append(m_me)
                resulting_data["m_sd"].append(m_sd)
                resulting_data["m_min"].append(m_min)
                resulting_data["m_max"].append(m_max)
                resulting_data["m_jc"].append(m_jc)
                resulting_data["m_jr"].append(m_jr)
                resulting_data["m_jv"].append(m_jv)
                resulting_data["m_iji"].append(m_iji)
                resulting_data["ns_me"].append(ns_me)
                resulting_data["ns_sd"].append(ns_sd)
                resulting_data["ns_sum"].append(ns_sum)
                resulting_data["ns_jc"].append(ns_jc)
                resulting_data["ns_jr"].append(ns_jr)
                resulting_data["ns_jv"].append(ns_jv)
                resulting_data["ns_iji"].append(ns_iji)
                resulting_data["nr_me"].append(nr_me)
                resulting_data["nr_sd"].append(nr_sd)
                resulting_data["nr_sum"].append(nr_sum)
                resulting_data["nr_jc"].append(nr_jc)
                resulting_data["nr_jr"].append(nr_jr)
                resulting_data["nr_jv"].append(nr_jv)
                resulting_data["nr_iji"].append(nr_iji)
                resulting_data["ir_me"].append(ir_me)
                resulting_data["ir_sd"].append(ir_sd)
                resulting_data["ax_bin"].append(ax_bin)
                resulting_data["ay_bin"].append(ay_bin)
                resulting_data["az_bin"].append(az_bin)
                resulting_data["gx_bin"].append(gx_bin)
                resulting_data["gy_bin"].append(gy_bin)
                resulting_data["gz_bin"].append(gz_bin)
                resulting_data["fa_bin"].append(fa_bin)
                resulting_data["fb_bin"].append(fb_bin)
                resulting_data["fc_bin"].append(fc_bin)
                resulting_data["fd_bin"].append(fd_bin)

                start += step
                i += 1
            0/0
        print("Writing results to {}.".format(file_name))
        df = DataFrame(resulting_data)
        df.to_csv(file_name, index=False)

    def _get_10_mag_bins(self, values, tim, window="hanning"):
        """
        Calculate 10 magnitude bins, based on frequencies
        """
        if not len(values):
            return [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        step = 5
        values_inter, tim_inter = self._interpolate(values, tim, step)
        if not values_inter:
            return [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

        # Frequency domain analysis
        n = len(values_inter)
        if not window:
            y = fft(values_inter)
            f = fftfreq(len(y))
        else:
            win = get_window(window, n)
            y = np.fft.rfft(values_inter * win)
            f = np.fft.rfftfreq(n)

        # Multiply frequency by sampling rate, to obtain Hertz
        f = f * (1000 / step)

        # Take only first half of the fft
        y = y[1 : len(y) // 2]
        f = f[1 : len(f) // 2]
        mag = [sqrt(r ** 2 + i ** 2) for r, i in zip(y.real, y.imag)]

        prev = 0
        step = 10
        f = list(f)
        squashed = {"freq": [], "value": []}
        for freq_step in range(step, 101, step):
            freq = [x for x in f if prev < x < freq_step]
            indices = [f.index(x) for x in freq]
            prev = freq_step
            squashed["freq"].append((prev + freq_step) / 2)
            squashed["value"].append(mean([mag[x] for x in indices]))
        return squashed["value"]

    @staticmethod
    def _interpolate(values, timestamps, step: int):
        """
        Interpolate list of sensor records, to the given milisecond interval.
        Return interpolated values and associated times.
        """
        current_step = timestamps[0] + timedelta(microseconds=step * 1000)
        tim = [datetime.timestamp(x) for x in timestamps]
        try:
            intepolator = interp1d(tim, values)
        except ValueError:
            return None, None
        values_inter = []
        tim_inter = []
        while current_step <= timestamps[-1]:
            values_inter.append(intepolator(datetime.timestamp(current_step)))
            tim_inter.append(current_step)
            current_step += timedelta(microseconds=step * 1000)
        return values_inter, tim_inter

    @staticmethod
    def _get_pir_data(data):
        """
        Structure pir data so it is ready for processing.
        """
        pir_data = {"sensor": [], "value": [], "timestamp": []}
        for sensor in ["pir_01", "pir_02", "pir_03", "pir_04", "pir_05", "pir_06"]:
            try:
                sub = data[sensor]
            except KeyError:
                continue
            pir_data["sensor"].extend([sensor] * len(sub))
            pir_data["value"].extend(list(sub["value"]))
            pir_data["timestamp"].extend(list(sub["timestamp"]))

        idx = argsort(array(pir_data["timestamp"]))
        pir_data["sensor"] = [pir_data["sensor"][i] for i in idx]
        pir_data["value"] = [pir_data["value"][i] for i in idx]
        pir_data["timestamp"] = [pir_data["timestamp"][i] for i in idx]

        return DataFrame(pir_data)

    @staticmethod
    def _process_pir_data(data, start, end, segment_interval, interval=1):
        """
        Classify every time step (determined by interval parameter) into one of categories:
        - moving (sensors, other than pir_01 active)
        - sitting (only pir_01 active)
        - nothing (no sensor activity)
        And calculate some features on top of this classes.
        """
        bins = []
        results = []
        while start < end:
            bins.append((start, start + timedelta(seconds=interval)))
            start += timedelta(seconds=interval)
        active = set()
        # Iterate through time steps
        for step in bins:
            activity = 0
            # Check for previously active sensors
            if active:
                activity = 2 if max(active) > 1 else 1
            if len(data["timestamp"]):
                # Iterate through each sensor record in a given time interval
                sub = data[
                    (data["timestamp"] >= step[0]) & (data["timestamp"] < step[1])
                ]
                for i in range(len(sub)):
                    if sub.iloc[i, :]["value"]:
                        active.add(int(sub.iloc[i, :]["sensor"][-1]))
                        if int(sub.iloc[i, :]["sensor"][-1]) > 1:
                            activity = 2
                        elif activity < 2:
                            activity = 1
                    else:
                        try:
                            active.remove(int(sub.iloc[i, :]["sensor"][-1]))
                        except KeyError:
                            # Removing what is already not present
                            pass
            # Add to results
            results.append(activity)

        #         # Fill parts where there are zeros between ones - the subject was not moving enough, but still sitting
        #         prev = data[0]
        #         for l in range(1, len(data)):
        #             if prev == 1 and data[i] == 0:
        #                 j = l
        #                 while data[j] == 0:
        #                     j += 1
        #                     if j + 1 >= len(data):
        #                         j -= 1
        #                         break
        #                 if data[j] == 1:
        #                     for k in range(l, j):
        #                         data[k] = 1
        #             prev = data[l]

        #         # Count number of intervals with specific class
        #         class_interval_count = [0, 0, 0]
        #         lengths = []
        #         prev = data[0]
        #         curr_len = 1
        #         for x in data[1:]:
        #             if x != prev:
        #                 class_interval_count[prev] += 1
        #                 if curr_len > 0:
        #                     lengths.append(curr_len)
        #                     curr_len = 1
        #             else:
        #                 curr_len += 1
        #             prev = x
        #         class_interval_count[data[-1]] += 1
        #         lengths.append(curr_len)
        return mean(results), std(results)

    @staticmethod
    def _process_signal(values):
        """
        Take Django query and do basic signal processing.
        """
        m = mean(values)
        s = std(values)

        return values, m, s

    @staticmethod
    def _join_accelerometer_signals(x, y, z):
        """
        Join accelerometer signals, based simply on concurrence.
        We can do this, as only one controller sends data in loop for all axis.
        """
        result = []
        n = min(len(x), len(y), len(z))
        for a, b, c in zip(x[:n], y[:n], z[:n]):
            result.append(sqrt(a ** 2 + b ** 2 + c ** 2))
        return result, mean(result), std(result)

    @staticmethod
    def _mean_crossing_rate(signal, m):
        """
        Calculate mean crossing rate from signal.
        Rate of mean crossings vs. the signal length.
        """
        try:
            prev = signal[0]
        except IndexError:
            return 0
        crosses = 0
        length = len(signal) - 1

        for curr in signal[1:]:
            if prev <= m < curr or prev > m >= curr:
                crosses += 1
            prev = curr
        if length < 1:
            return 0
        return crosses / length

    @staticmethod
    def _mean_acceleration_intensity(signal):
        """
        Mean derivative of a signal.
        """
        try:
            prev = signal[0]
        except IndexError:
            return 0
        length = len(signal) - 1
        derv = []

        for curr in signal[1:]:
            derv.append(abs(curr - prev))
            prev = curr

        return mean(derv)

    @staticmethod
    def _join_cpu_signals(a, b, c, d):
        """
        Similar to accelerometer one.
        """
        result = []
        n = min(len(a), len(b), len(c), len(d))
        for w, x, y, z in zip(a[:n], b[:n], c[:n], d[:n]):
            result.append(sqrt(w ** 2 + x ** 2 + y ** 2 + z ** 2))
        return result, mean(result), std(result)

    def _get_cpu_stats(self, val):
        if not val:
            return 0, 0, 0
        return min(val), max(val), self._mean_crossing_rate(val, mean(val))

    @staticmethod
    def _find_ram_jump(signal):
        if not signal:
            return [], {}
        derivative = []
        prev = signal[0]
        for curr in signal[1:]:
            derivative.append(abs(curr - prev))
            prev = curr
        peaks, _ = find_peaks(derivative, threshold=0.25)

        p = {"position": [], "magnitude": []}
        for x in peaks:
            p["position"].append(x)
            p["magnitude"].append(derivative[x])
        return derivative, p

    @staticmethod
    def _get_mem_stats(val, peaks, derivatives):
        # Calculate average inter jump interval
        intervals = []
        if peaks and peaks["position"]:
            prev = peaks["position"][0]
            for curr in peaks["position"][1:]:
                intervals.append(curr - prev)
                prev = curr
        if val:
            avg_load = round(mean(val), 2)
            min_load = min(val)
            max_load = max(val)
        else:
            avg_load = 0
            min_load = 0
            max_load = 0
        if peaks:
            jump_count = len(peaks["position"])
            if derivatives:
                jump_rate = round(len(peaks["position"]) / len(derivatives), 2)
            else:
                jump_rate = 0
            avg_jump_value = round(mean(peaks["magnitude"]), 2)
            avg_inter_jump_interval = round(mean(intervals), 2)
        else:
            jump_count = 0
            jump_rate = 0
            avg_jump_value = 0
            avg_inter_jump_interval = 0
        return (
            avg_load,
            min_load,
            max_load,
            jump_count,
            jump_rate,
            avg_jump_value,
            avg_inter_jump_interval,
        )

    @staticmethod
    def _find_net_jump(signal):
        if not signal:
            return [], {}
        derivative = []
        prev = signal[0]
        for curr in signal[1:]:
            derivative.append(abs(curr - prev))
            prev = curr
        peaks, _ = find_peaks(derivative, threshold=mean(derivative))

        p = {"position": [], "magnitude": []}
        for x in peaks:
            p["position"].append(x)
            p["magnitude"].append(derivative[x])
        return derivative, p

    @staticmethod
    def _get_net_stats(val, peaks, derivatives):
        # Calculate average inter jump interval
        intervals = []
        if peaks and peaks["position"]:
            prev = peaks["position"][0]
            for curr in peaks["position"][1:]:
                intervals.append(curr - prev)
                prev = curr
        try:
            sum_load = val[-1] - val[0]
        except IndexError:
            sum_load = 0
        if peaks:
            jump_count = len(peaks["position"])
            if derivatives:
                jump_rate = round(len(peaks["position"]) / len(derivatives), 2)
            else:
                jump_rate = 0
            avg_jump_value = round(mean(peaks["magnitude"]), 2)
            avg_inter_jump_interval = round(mean(intervals), 2)
        else:
            jump_count = 0
            jump_rate = 0
            avg_jump_value = 0
            avg_inter_jump_interval = 0
        return sum_load, jump_count, jump_rate, avg_jump_value, avg_inter_jump_interval

    # LOADING CSV DATA
    def load_csv_data(self):
        """
        Load data from csv files and generate datasets with given time steps.
        Calculate additional frequency features.
        """

        datasets = []
        task_intervals = []
        for t in settings.TASKS:
            for s in settings.SEGMENT_INTERVALS:
                file_name = settings.DATA_DIR + settings.CSV_TEMPLATE.format(t, s)
                try:
                    datasets.append(read_csv(file_name).fillna(0))
                    task_intervals.append((t, s))
                except FileNotFoundError:
                    print("File {} missing.".format(file_name))

        # Calculate magnitude bins
        for i, _ in enumerate(datasets):
            for col in [
                "ax_bin",
                "ay_bin",
                "az_bin",
                "gx_bin",
                "gy_bin",
                "gz_bin",
                "fa_bin",
                "fb_bin",
                "fc_bin",
                "fd_bin",
            ]:
                # Transform string of a list into a list of floats
                bins = array(
                    [
                        [float(y) for y in x.strip("[]").split(",")]
                        for x in datasets[i][col]
                    ]
                ).T
                for j in range(len(bins)):
                    datasets[i]["{}_{}".format(col, j)] = bins[j]
                datasets[i] = datasets[i].drop(col, axis=1)
                datasets[i]["{}_max".format(col)] = np.max(bins, axis=0)
                datasets[i]["{}_min".format(col)] = np.min(bins, axis=0)
                datasets[i]["{}_mean".format(col)] = mean(bins, axis=0)
                datasets[i]["{}_std".format(col)] = std(bins, axis=0)
        datasets = [x.fillna(0) for x in datasets]

        results = {}
        task = 0
        for i, data in enumerate(datasets):
            results.update(
                {
                    "{}_{}".format(
                        str(task_intervals[i][0]).zfill(2),
                        str(task_intervals[i][1]).zfill(3),
                    ): data
                }
            )
            task += 1

        return results
