import re

from pandas import read_csv, unique, to_datetime


class DataSetHandler:
    def __init__(self, file_name):
        self.file_name = file_name

        self.dataframe = None
        # Statistics extracted from the dataframe
        self.users = None
        self.seances = None
        self.user_seances = None
        self.sensors = None
        self.start_ends = None

    def import_data(self):
        """
        Import a given csv file into a pandas dataframe. Also extract some
        basic features of the data and store them in class attributes.
        """
        print("Importing dataframe from file {}.".format(self.file_name))
        self.dataframe = read_csv(self.file_name)

        timestamps = self._str_to_datetime(list(self.dataframe.timestamp))
        self.dataframe = self.dataframe.assign(timestamp=timestamps)

        self.dataframe["timestamp"] = to_datetime(
            self.dataframe["timestamp"], format="%Y-%m-%d %H:%M:%S.%f+02"
        )
        self._update_stats()
        print("Dataframe imported successfully.")

    def subset_data(self, seance_ids):
        """
        Subset the dataframe, according to the given seance ids.
        """
        print()
        print(
            "Subsetting the dataframe, removing {} seances.".format(
                len(self.seances) - len(seance_ids)
            )
        )
        self.dataframe = self.dataframe[self.dataframe["seance_id"].isin(seance_ids)]
        self._update_stats()
        print("Dataframe subset successfully.")

    def _update_stats(self):
        """
        Update various stats, usually after data subsampling.
        """
        self.users = list(unique(self.dataframe["user_id"]))
        self.seances = list(unique(self.dataframe["seance_id"]))
        self.sensors = list(unique(self.dataframe["sensor"]))
        self.user_seances = sorted(
            list(
                set(
                    [
                        x
                        for x in zip(
                            self.dataframe["user_id"], self.dataframe["seance_id"]
                        )
                    ]
                )
            )
        )

        self.start_ends = {}
        starts = self._str_to_datetime(list(self.dataframe.start))
        ends = self._str_to_datetime(list(self.dataframe.end))
        for x in unique(
            [
                (id, (start, end))
                for id, start, end in zip(self.dataframe["seance_id"], starts, ends,)
            ]
        ):
            self.start_ends.update({x[0]: (x[1][0], x[1][1])})

    @staticmethod
    def _str_to_datetime(timestamps):
        """
        Match and fix the datetime strings based on
        regex and return a list of datetime objects.
        """
        results = []
        for x in timestamps:
            result = re.search("[0-9]{2}:[0-9]{2}:[0-9]{2}?:.|[0-9]{0,5}\+02", x)
            if result:
                start = 20
                end = result.end() - 3
                if "." not in x:
                    x = "".join(
                        [x[: start - 1], ".", x[start:end].ljust(6, "0"), x[end:]]
                    )
                else:
                    x = "".join([x[:start], x[start:end].ljust(6, "0"), x[end:]])
            results.append(x)
        return results

    def get_data_for_seance(self, seance_id):
        """
        Return the seance subset of the dataframe.
        """
        return self.dataframe[self.dataframe["seance_id"] == seance_id]

    def is_2_seances(self, seance_id):
        """
        Check if user completed at least 2 seances of this task.
        """
        user_id = [x for x in self.user_seances if x[1] == seance_id][0][0]
        return len([x for x in self.user_seances if x[0] == user_id]) >= 2
